﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.ScatterSearch.TravelingSalesman {
  /// <summary>
  /// An operator that relinks paths between traveling salesman solutions.
  /// </summary>
  [Item("TravelingSalesmanPathRelinker", "An operator that relinks paths between traveling salesman solutions.")]
  [StorableClass]
  public sealed class TravelingSalesmanPathRelinker : PathRelinker {
    [StorableConstructor]
    private TravelingSalesmanPathRelinker(bool deserializing) : base(deserializing) { }
    private TravelingSalesmanPathRelinker(TravelingSalesmanPathRelinker original, Cloner cloner) : base(original, cloner) { }
    public TravelingSalesmanPathRelinker()
      : base() {
      ParentsParameter.ActualName = "TSPTour";
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TravelingSalesmanPathRelinker(this, cloner);
    }

    public static ItemArray<IItem> Apply(IItem initiator, IItem guide, PercentValue n) {
      var v1 = initiator.Clone() as Permutation;
      var v2 = guide as Permutation;

      if (v1.Length != v2.Length)
        throw new ArgumentException("TravelingSalesmanPathRelinker: The solutions are of different length.");
      if (n.Value == 0.0)
        throw new ArgumentException("TravelingSalesmanPathRelinker: RelinkingAccuracy is 0.");

      var solutions = new List<Permutation>();
      for (int i = 0; i < v1.Length; i++) {
        if (v1[i] != v2[i]) {
          var target = v1.Select((x, index) => new { Value = x, ValueIndex = index }).First(x => x.Value == v2[i]);
          // XOR swap
          v1[i] ^= v1[target.ValueIndex];
          v1[target.ValueIndex] ^= v1[i];
          v1[i] ^= v1[target.ValueIndex];
          solutions.Add((Permutation)v1.Clone());
        }
      }

      var stepSize = 1 / n.Value;
      var selection = new List<IItem>();
      for (int i = 0; i < solutions.Count; i = (int)Math.Round(i + stepSize)) {
        selection.Add(solutions.ElementAt(i));
      }

      return new ItemArray<IItem>(selection);
    }

    protected override ItemArray<IItem> Relink(ItemArray<IItem> parents, PercentValue n) {
      if (parents.Length != 2) throw new ArgumentException("TravelingSalesmanPathRelinker: The number of parents is not equal to 2.");
      return Apply(parents[0], parents[1], n);
    }
  }
}
